#lang racket
(require plot)
(require text-table)

(define (data-path symbol period)
  (string-append "StockHistory/" period "/" symbol ".csv"))

(struct hist (date open high low close adj-close vol)
  #:transparent)

(define symbols '("^DJI" "^GSPC" "AAPL" "AMZN"
                         "BAC" "F" "MSFT" "VBMFX" "XOM"))

(define symbol-color
  (make-hash
   (list
    (cons "^DJI" "black")
    (cons "^GSPC" "gray")
    (cons "AAPL" "black")
    (cons "AMZN" "gray")
    (cons "BAC" "purple")
    (cons "F" "orange")
    (cons "MSFT" "blue")
    (cons "VBMFX" "black")
    (cons "XOM" "gray")
    )))

(define symbol-style
  (make-hash
   (list
    (cons "^DJI" 'solid)
    (cons "^GSPC" 'solid)
    (cons "AAPL" 'dot)
    (cons "AMZN" 'dot)
    (cons "BAC" 'dot-dash)
    (cons "F" 'solid)
    (cons "MSFT" 'long-dash)
    (cons "VBMFX" 'short-dash)
    (cons "XOM" 'short-dash)
    )))

(define (import-record port)
  (let ([rec (read-line port)])
    (if (eof-object? rec)
        eof
        (match (string-split rec ",")
          [(list date open high low close adj-close vol)
           (hist date
                 (string->number open)
                 (string->number high)
                 (string->number low)
                 (string->number close)
                 (string->number adj-close)
                 (string->number vol))]
          [_ (error "Failed to load record.")]))))


(define month-days
  #(0 0 31 59.25 90.25 120.25 151.25 181.25
      212.25 243.25 273.25 304.25 334.25))

(define (date->number d)
  (match (string-split d "-")
    [(list year month day)
     (let ([year (string->number year)]
           [month (string->number month)]
           [day (string->number day)])
       (exact->inexact (+ year
                          (/ (vector-ref month-days month) 365.25)
                          (/ (sub1 day) 365.25))))]))

(define (get-coords symbol period filter normalize)
  (let ([in-port (open-input-file (data-path symbol period))]
        [start-price #f])
    (read-line in-port)
    (let* ([recs
            (reverse
             (let loop([rec-list '()])
               (let ([rec (import-record in-port)])
                 (if (eof-object? rec)
                     rec-list
                     (if (filter rec)
                         (let ([date-val (date->number (hist-date rec))]
                               [close (hist-close rec)])
                           (unless start-price (set! start-price close))
                           (let ([val (if normalize (/ close start-price) close)])
                             (loop (cons (vector date-val val) rec-list))))
                         (loop rec-list))))))])
      (close-input-port in-port)
      recs)))

(define (plot-symbols symbols period filter
                      [normalize #f]
                      [anchor 'top-left]
                      [y-min #f]
                      [y-max #f])
  (let* ([plot-data
          (for/list ([symbol symbols])
            (let ([color (hash-ref symbol-color symbol)]
                  [style (hash-ref symbol-style symbol)])
              (lines (get-coords symbol period filter normalize)
                     #:label symbol
                     #:color color
                     #:width 1.5
                     #:style style)))]
         [ymin (if (and normalize (not y-min)) 0.0 y-min)]
         [ymax (if (and normalize (not y-max)) 2.0 y-max)])
    (parameterize
        ([plot-width 400]
         [plot-height 250]
         [plot-x-label "Year"]
         [plot-y-label #f]
         [plot-legend-anchor anchor])
      (plot plot-data
            #:y-min ymin
            #:y-max ymax))))

(define (get-data symbol period filter field)
  (let ([in-port (open-input-file (data-path symbol period))])
    (read-line in-port)
    (let* ([recs
            (reverse
             (let loop([rec-list '()])
               (let ([rec (import-record in-port)])
                 (if (eof-object? rec)
                     rec-list
                     (if (filter rec)
                         (loop (cons (hist-close rec) rec-list))
                         (loop rec-list))))))])
      (close-input-port in-port)
      recs)))

(define (bin-table bins bin-min bin-max)
  (let* ([num-bins (length bins)]
         [bin-delta (/ (- bin-max bin-min) num-bins)]
         [rows
          (for/list ([i (in-range num-bins)]
                     [bin bins])
            (let ([bmin (+ bin-min (* bin-delta i))]
                  [bmax (+ bin-min (* bin-delta (add1 i)))]
                  [count (vector-ref bin 1)])
              (list
               (~r i #:min-width 3)
               (~r bmin #:min-width 8  #:precision (list '= 2))
               (~r bmax #:min-width 8  #:precision (list '= 2))
               (~r count #:min-width 4))))])
    (table->string (cons '("Bin" "   Min" "   Max" "Vals") rows))))

#;(define (get-summary data)
    (let ([min-value 99999999999]
          [max-value 0]
          [sum 0])
      (let loop ([val-list data])
        (unless (null? val-list)
          (let ([val (car val-list)])
            (set! sum (+ sum val))
            (when (< val min-value)
              (set! min-value val))
            (when (> val max-value)
              (set! max-value val))
            (loop (cdr val-list)))))
      (values min-value max-value (/ sum (length data)))))

(define (categorize data avg num-bins)
  (let* ([bin (make-vector num-bins 0)]
         [bin-min (* 0.4 avg)]
         [bin-max (* 1.6 avg)]
         [bin-delta (/ (- bin-max bin-min) num-bins)])
    (define (update-bin val)
      (when (<= bin-min val bin-max)
        (let ([i (inexact->exact (floor (/ (- val bin-min) bin-delta)))])
          (vector-set! bin i (add1 (vector-ref bin i))))))
    (let loop ([val-list data])
      (unless (null? val-list)
        (update-bin (car val-list))
        (loop (cdr val-list))))
    (values bin-min bin-max
            (for/list ([i (in-range num-bins)])
              (vector i (vector-ref bin i))))))

(define (histogram-symbol symbol period filter [bins 11])
  (let*-values ([(data) (get-data symbol period filter hist-close)]
                [(avg) (/ (apply + data) (length data))]
                [(bin-min bin-max hist-data) (categorize data avg bins)])
    (displayln (bin-table hist-data bin-min bin-max))
    (parameterize
        ([plot-width 400]
         [plot-height 250]
         [plot-x-label #f]
         [plot-y-label "Frequency"])
      (plot (discrete-histogram hist-data)))))

(define (year-range y1 y2)
  (lambda (rec) 
    (string<=? y1 (substring (hist-date rec) 0 4)
               y2)))

(define (date-range d1 d2)
  (lambda (rec) 
    (string<=? d1 (substring (hist-date rec) 0 10)
               d2)))

(define (int-rate v p t)
  (let ([n 12.0])
    (* n (- (expt (/ v p) (/ 1 (* n t))) 1))))

(define (std-deviation nums)
  (let* ([n (length nums)]
         [mu (/ (apply + nums) n)]
         [sqr-diff (map (lambda (x) (sqr (- x mu))) nums)])
    (sqrt (/ (apply + sqr-diff) n))))
               
(define (regression-params data)
  (define (x v) (vector-ref v 0))
  (define (y v) (vector-ref v 1))
  (let* ([num (length data)] 
         [totx (apply + (map x data))]
         [toty (apply + (map y data))]
         [avgx (/ totx num)]
         [avgy (/ toty num)]
         [ss-xy (apply + (map (lambda (v) (* (- (x v) avgx) (- (y v) avgy))) data))]
         [ss-xx (apply + (map (lambda (v) (sqr (- (x v) avgx))) data))]
         [b (/ ss-xy ss-xx)]
         [a (- avgy (* b avgx))])
    (values a b)))
         
(define (plot-regression symbol period filter
                         [anchor 'top-left])
  (let* ([coords (get-coords symbol period filter #f)]
         [plot-data 
          (let ([color (hash-ref symbol-color symbol)])
            (points coords #:label symbol #:color color))])
    (let-values ([(a b) (regression-params coords)])
      (parameterize 
          ([plot-width 400]
           [plot-height 250]
           [plot-x-label "Year"]
           [plot-y-label #f]
           [plot-legend-anchor anchor])
        (plot (list
               plot-data
               (function (lambda (x) (+ (* b x) a))
                         #:color "black" #:label "Regr")))))))

(define (correlation data)
  (define (x v) (vector-ref v 0))
  (define (y v) (vector-ref v 1))
  (let* ([num (length data)] 
         [totx (apply + (map x data))]
         [toty (apply + (map y data))]
         [avgx (/ totx num)]
         [avgy (/ toty num)]
         [ss-xx (apply + (map (lambda (v) (sqr (- (x v) avgx))) data))]         
         [ss-yy (apply + (map (lambda (v) (sqr (- (y v) avgy))) data))]
         [ss-xy (apply + (map (lambda (v) (* (- (x v) avgx) (- (y v) avgy))) data))]
         [b (/ ss-xy ss-xx)]
         [a (- avgy (* b avgx))]
         [ssr (apply + (map (lambda (v) (sqr (- (+ (* b (x v)) a) avgy))) data))])
    (/ ssr ss-yy)))
