#lang racket
(require data/heap)

; priority queue and helper functions

(define (comp n1 n2)
  (let ([d1 (cdr n1)]
        [d2 (cdr n2)])
    (<= d1 d2)))

(define queue (make-heap comp))

(define (enqueue n) (heap-add! queue n))

(define (dequeue)
  (let ([n (heap-min queue)])
    (heap-remove-min! queue)
    n))
            
; Puzzle Structure ------------------------------------------

(define SIZE 4)
(define empty (sqr SIZE))
  
(define (ref puzzle r c)
  (let ([i (+ c (* r SIZE))])
    (vector-ref puzzle i)))

(define tile-homes
  (let ([hash (make-hash)])
    (for ([n (in-range (sqr SIZE))])
      (hash-set! hash (add1 n) (cons (quotient n SIZE) (remainder n SIZE))))
    hash))

(define (cost puzzle guard)
  (let ([dist 0])
    (for* ([r SIZE] [c SIZE])
      (let ([t (ref puzzle r c)])
        (when (guard r c t)
          (let* ([th (hash-ref tile-homes t)]
                 [hr (car th)]
                 [hc (cdr th)]
                 [d (+ (abs (- hr r)) (abs (- hc c)))])
            (set! dist (+ dist d))))))
    dist))

(define (manhattan puzzle)
  (cost puzzle (lambda (r c t) 
                 (not (= t empty)))))
  
(define (zone1a puzzle)
  (cost puzzle (lambda (r c t)
                 (or (and (<= t 4) (> r 1))))))
                 
(define (zone1b puzzle)
  (cost puzzle (lambda (r c t)
                 (<= t 4))))

(define zone2-tiles (set 5 9 13))

(define (zone2 puzzle)
  (cost puzzle (lambda (r c t)
                 (and (>= r 1)
                      (set-member? zone2-tiles t)))))


(define (move-offset i)
  (case i
    [(0) (values  0 -1)]
    [(1) (values  0  1)] 
    [(2) (values -1  0)] 
    [(3) (values  1  0)]))

(define (empty-loc puzzle)
  (values
   (vector-ref puzzle empty)
   (vector-ref puzzle (add1 empty))))

(define (make-move puzzle i)
  (let*-values ([(ro co) (move-offset i)]
                [(re ce) (empty-loc puzzle)]
                [(rt ct) (values (+ re ro) (+ ce co))]
                [(t) (ref puzzle rt ct)])
    (for/vector ([i (in-range (+ 2 (sqr SIZE)))])
      (cond [(< i empty)
             (let-values ([(r c) (quotient/remainder i SIZE)])
               (cond [(and (= r re) (= c ce)) t]
                     [(and (= r rt) (= c ct)) empty]
                     [else (vector-ref puzzle i)]))]
            [(= i empty) rt]
            [else ct]))))

(define min-r 0)
(define min-c 0)

(define (next-states puzzle)
  (let-values ([(re ce) (empty-loc puzzle)])
    (define (legal i)
      (let*-values ([(ro co) (move-offset i)]
                    [(rt ct) (values (+ re ro) (+ ce co))])
        (and (>= rt min-r) (>= ct min-c) (< rt SIZE) (< ct SIZE))))
    (for/list ([i (in-range 4)] #:when (legal i))
      (make-move puzzle i))))
       
(define (print puzzle)
  (for* ([r SIZE] [c SIZE])
    (when (= 0 c) (printf "\n"))
    (let ([t (ref puzzle r c)])
      (if (= t empty)
          (printf "   ")
          (printf " ~a" (~a t #:min-width 2 #:align 'right) ))))
  (printf "\n"))

; ---------------------------------------------------
(define closed (make-hash))

(define (solution-list n l)
  (if (equal? n null)
      l
      (let* ([parent (first (hash-ref closed n))])
        (solution-list parent (cons n l)))))

(define (print-solution l)
  (for ([p l]) (print p)))

(define (process-closed node-parent node node-depth score fscore)
  (begin
    (hash-set! closed node (list node-parent score)) 
    (for ([child (next-states node)])
      (let* ([depth (add1 node-depth)]
             [next-score (+ depth (fscore child))]
             [next (cons (list child depth node) next-score)])
        (if (hash-has-key? closed child)
            (let* ([prior-score (second (hash-ref closed child))])
              (when (< next-score prior-score)
                (hash-remove! closed child)
                (enqueue next)))
            (enqueue next))))))

(define (a-star puzzle fscore)
  (let ([solution null]
        [goal null])
    (hash-clear! closed)
    (set! queue (make-heap comp))  ; open
    (enqueue (cons (list puzzle 0 null) (fscore puzzle)))
    (let loop ()
      (when (equal? solution null)
        (let* ([node-info (dequeue)])
          (match node-info
            [(cons (list node node-depth node-parent) score)
             (if (= 0 (fscore node))
                 (begin
                   (set! goal node)
                   (set! solution (solution-list node-parent (list node))))
                 (process-closed node-parent node node-depth score fscore))])
          (if (> (heap-count queue) 0)
              (loop)
              (when (equal? solution null) (printf "No solution found\n"))))))
    (values goal solution)))


(define (solve puzzle)  
  (set! min-r 0)
  (set! min-c 0)
  (let*-values ([(goal sol-z1a) (a-star puzzle zone1a)])
    ;(print-solution sol1)
    (let*-values  ([(goal sol2) (a-star goal zone1b)]
                   [(sol-z1b) (cdr sol2)])
      ;(print-solution sol2)
      (set! min-r 1)
      (let*-values  ([(goal sol3) (a-star goal zone2)]
                     [(sol-z2) (cdr sol3)])
        ;(print-solution sol3)
        (set! min-c 1)
        (let*-values  ([(goal sol4) (a-star goal manhattan)]
                       [(sol-man) (cdr sol4)])
          ;(print-solution sol4)
          (print-solution (append sol-z1a sol-z1b sol-z2 sol-man)))))))

;(solve #(7 8 12 5 3 13 1 9 2 10 11 6 14 15 4 16 3 3))
;(solve #(2 10 8 3 1 6 16 4 5 9 7 11 13 14 15 12 1 2))
;(solve #(5 1 2 4 14 9 3 7 13 10 12 6 15 11 8 16 3 3))
;(solve #(15 2 1 12 8 5 6 11 4 9 10 7 3 13 14 16 3 3))
;(solve #(10 9 5 13 8 14 15 7 1 3 11 6 4 2 12 16 3 3))
;(solve #(3 1 2 4 13 6 7 8 5 12 10 11 9 14 15 16 3 3))
;(solve #(9 6 12 3 5 13 16 8 14 1 10 7 2 15 11 4 1 2))
;(solve #(11 1 3 12 5 2 9 8 10 6 14 15 7 13 4 16 3 3))
;(solve #(1 6 15 9 7 2 13 8 4 10 11 12 3 14 5 16 3 3))
;(solve #(5 11 12 4 2 6 7 14 9 8 15 1 13 10 3 16 3 3))
;(solve #(7 3 6 2 4 1 13 15 9 14 11 5 12 10 8 16 3 3))
;(solve #(4 12 11 1 6 2 7 15 13 9 10 5 3 8 14 16 3 3))
;(solve #(10 2 15 11 13 9 6 3 5 14 12 4 1 8 7 16 3 3)) 
(solve #(1 8 12 7 4 6 2 10 9 14 3 11 13 5 15 16 3 3))
;(solve #(4 2 15 8 13 7 10 14 1 5 11 12 3 9 6 16 3 3))
;(solve #(10 14 13 4 2 6 7 12 8 3 9 11 15 1 5 16 3 3))
;(solve #(9 2 4 15 6 3 13 8 1 7 5 14 11 12 10 16 3 3))
;(solve #(9 2 3 14 4 5 6 7 12 11 10 13 1 15 8 16 3 3))
